#include <cstdio>
#include <cstdarg>

#include "level.hh"
#include "defs.hh"
#include "rom.hh"

void Level::WritePalette(unsigned addr) const
{
    for(unsigned a=0; a<16*3; ++a)
        ROM[addr+a] = Palette[a];
}

void Level::WriteActives(unsigned first_screen, unsigned addr, unsigned end_addr) const
{
    unsigned n=0;
    for(unsigned a=0; a<MaxActives; ++a)
    {
        if(actives[a].screen == 255) continue;
        unsigned begin = addr + 1 + n*6;
        if(begin+6 > end_addr)
        {
            fprintf(stderr, "Error: Too many actives!\n");
        }
        ROM[begin+0] = actives[a].type;
        ROM[begin+1] = actives[a].screen + first_screen;
        ROM[begin+2] = actives[a].x1;
        ROM[begin+3] = actives[a].y1;
        ROM[begin+4] = actives[a].x2;
        ROM[begin+5] = actives[a].y2;
        ++n;
    }
    ROM[addr] = n + 1;
    for(unsigned a = addr + 1 + n*6; a < end_addr; ++a)
        ROM[a] = 0xFF;
}

void Level::WriteEnemies
    (unsigned first_screen,
     unsigned ptraddr, unsigned& addr, unsigned end_addr) const
{
    ROM[ptraddr+0] = ((addr|0x8000) & 0x00FF);
    ROM[ptraddr+1] = ((addr|0x8000) & 0xFF00) >> 8;
    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        if(enemies[a].screen == 255)
        {
            if(addr+1 > end_addr)
            {
                fprintf(stderr, "Error: Too many enemies (can't write end byte)!\n");
            }
            ROM[addr+0] = 0xFF;
            addr += 1;
            break;
        }
        else
        {
            if(addr+4 > end_addr)
            {
                fprintf(stderr, "Error: Too many enemies (%X,%X)!\n", addr,end_addr);
            }
            
            ROM[addr+0] = enemies[a].screen;
            if(enemies[a].screen != 128) ROM[addr+0] += first_screen;
            ROM[addr+1] = enemies[a].x;
            ROM[addr+2] = enemies[a].y;
            ROM[addr+3] = enemies[a].type;
            addr += 4;
        }
    }
}

void Level::ReadPalette(unsigned addr)
{
    for(unsigned a=0; a<16*3; ++a)
        Palette[a] = ROM[addr+a];
}

void Level::ReadActives(unsigned first_screen, unsigned addr)
{
    unsigned n = ROM[addr++];
    if(n > MaxActives)n=MaxActives;
    for(unsigned a=0; a<MaxActives; ++a) actives[a].Delete();
    for(unsigned a=0; a<n; ++a)
    {
        if(ROM[addr+a*6+1] == 255) break;
        //fprintf(stderr, "Reading active from %X\n", addr);
        actives[a].type   = ROM[addr+a*6+0];
        actives[a].screen = ROM[addr+a*6+1] - first_screen;
        actives[a].x1     = ROM[addr+a*6+2];
        actives[a].y1     = ROM[addr+a*6+3];
        actives[a].x2     = ROM[addr+a*6+4];
        actives[a].y2     = ROM[addr+a*6+5];
    }
}

void Level::ReadEnemies(unsigned first_screen, unsigned ptraddr)
{
    unsigned data_addr = ROM[ptraddr+0] | (ROM[ptraddr+1]<<8);
    data_addr = (ptraddr & 0x1C000) + (data_addr & 0x3FFF);
    //printf("// Enemies %04X -> %06X\n", ptraddr, data_addr);
    for(unsigned a=0; a<MaxEnemies; ++a) enemies[a].Delete();
    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        if(ROM[data_addr+a*4] == 255) break;
        /*
        printf("// enemy %u @ %X: %02X %02X %02X %02X\n",
          a, data_addr+a*4,
          ROM[data_addr+a*4+0],
          ROM[data_addr+a*4+1],
          ROM[data_addr+a*4+2],
          ROM[data_addr+a*4+3]);
        */
        enemies[a].screen = ROM[data_addr+a*4+0];
        if(enemies[a].screen != 128) enemies[a].screen -= first_screen;
        enemies[a].x      = ROM[data_addr+a*4+1];
        enemies[a].y      = ROM[data_addr+a*4+2];
        enemies[a].type   = ROM[data_addr+a*4+3];
    }
}

void Level::DeleteGutsBlockActives()
{
    /* Gutsblocks. */
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == Last) break;
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
              if(sc.ascii[y*32+x] == Ascii_Guts)
                  for(unsigned a=0; a<MaxActives; ++a)
                  {
                      Active& act = actives[a];
                      if(act.screen == s
                      && act.type == 2
                      && act.x1 == x*16
                      && act.y1 == y*16
                      && act.x2 == (((x+2)*16) & 0xFF)
                      && act.y2 == (((y+2)*16) & 0xFF))
                      {
                          act.Delete();
                      }
                  }
    }
}

void Level::MakeAsciiActives()
{
    unsigned buzzcounter[MaxRooms] = {0};
    
    // Buzzblocks are ordered last-to-first.
    // Thus we have to walk the structure backwards.
    for(unsigned a=MaxActives; a-- > 0; )
    {
        Active& act = actives[a];
        if(act.screen == 255) continue;
        if(act.type == 1 && act.y2 <= act.y1)
        {
            // flying fire
            unsigned x1 = act.x1; x1 /= 16;
            unsigned x2 = act.x2; x2 /= 16;
            unsigned y = act.y1; y /= 16;
            for(unsigned x=x1; x<x2; ++x)
                rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjFire;
            
            act.Delete();
        }
        if(act.type == 1 && act.y2 > act.y1)
        {
            // leftside shutter
            unsigned x1 = act.x1; x1 /= 16;
            unsigned y1 = act.y1; y1 /= 16;
            unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
            unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
            
            for(unsigned y=y1; y<y2; ++y)
                for(unsigned x=x1; x<x2; ++x)
                    rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjShutterIn;
            
            act.Delete();
        }
        if(act.type == 3)
        {
            // rightside shutter
            unsigned x1 = act.x1; x1 /= 16;
            unsigned y1 = act.y1; y1 /= 16;
            unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
            unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
            
            for(unsigned y=y1; y<y2; ++y)
                for(unsigned x=x1; x<x2; ++x)
                    rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjShutterOut;
            
            act.Delete();
        }

        if(act.type == 0 && false)
        {
            // buzz (appearing and disappearing blocks)
            unsigned x1 = act.x1; x1 /= 16;
            unsigned y1 = act.y1; y1 /= 16;
            unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
            unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
            
            for(unsigned y=y1; y<y2; ++y)
                for(unsigned x=x1; x<x2; ++x)
                {
                    unsigned& count = buzzcounter[act.screen];
                    if(count > 9)
                    {
                        fprintf(stderr, "Too many buzzblocks in one screen!\n");
                        break;
                    }
                    rooms[act.screen].ascii[y*32+16+x] = "0123456789"[count++];
                }
            
            act.Delete();
        }
    }
}

void Level::MakeAsciiEnemies()
{
    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        Enemy& e = enemies[a];
        if(e.screen == 255) continue;
        
        char ch = 0;
        if(e.type < sizeof(AsciiObjects)/sizeof(*AsciiObjects))
            ch = AsciiObjects[e.type].ascii;
        
        if(ch != 0)
        {
            unsigned x = e.x / 16;
            unsigned y = e.y / 16;
            
            char& target = rooms[e.screen].ascii[y*32+16+x];
            if(target == ' ')
            {
                target = ch;
                e.Delete();
            }
        }
    }
}

void Level::CreateGutsBlockActives()
{
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == Last) break;
        char ignore[16*16] = {0};
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
            {
                if(ignore[y*16+x]) continue;
                if(sc.ascii[y*32+x] == Ascii_Guts)
                {
                    AddActive(Active(2, s, x*16,y*16, (x+2)*16, (y+2)*16));
                    ignore[y*16+x+1]=1;
                    ignore[(y+1)*16+x]=1;
                    ignore[(y+1)*16+x+1]=1;
                }
            }
    }
}

void Level::ReadAsciiActives()
{
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == Last) break;
        char ignore[16*16] = {0};
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
            {
                if(ignore[y*16+x]) continue;
                char c = sc.ascii[y*32+16+x];
                switch(c)
                {
                    case Ascii_ObjShutterIn:
                    case Ascii_ObjShutterOut:
                    {
                        unsigned y1=y;
                        unsigned y2=y;
                        while(y2<16 && sc.ascii[y2*32+16+x]==c) ++y2;
                        unsigned x1=x;
                        unsigned x2=x;
                        while(x2<16 && sc.ascii[y1*32+16+x2]==c) ++x2;
                        
                        fprintf(stderr, "Creating shutter %c %u,%u,%u,%u\n",
                            c,x1,y1,x2,y2);
                        
                        int t = c==Ascii_ObjShutterIn ? 1 : 3;
                        AddActive(Active(t, s, x1*16,y1*16, x2*16,y2*16));

                        for(unsigned Y=y1; Y<y2; ++Y)
                            for(unsigned X=x1; X<x2; ++X)
                                ignore[Y*16+X]=1;
                        break;
                    }
                    case Ascii_ObjFire:
                    {
                        AddActive(Active(1, s, x*16, y*16, (x+1)*16, y*16));
                        break;
                    }
                }
            }
    }
}

void Level::ReadAsciiEnemies()
{
    const unsigned nobj = sizeof(AsciiObjects)/sizeof(*AsciiObjects);
    
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == Last) break;
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
            {
                char c = sc.ascii[y*32+16+x];
                if(c==' ') continue;
                
                for(unsigned a=0; a<nobj; ++a)
                    if(c == AsciiObjects[a].ascii)
                    {
                        c=0;
                        AddEnemy(Enemy(s, x*16 + AsciiObjects[a].x,
                                          y*16 + AsciiObjects[a].y,
                                          a));
                        break;
                    }
                    
                if(c)
                {
                    fprintf(stderr, "Unrecognized object '%c'\n", c);
                }
            }
    }
}

void GenericAsciiMaker(Screen& s, int n,...)
{
    char map[0x100*4];
    for(unsigned a=0; a<sizeof(map); ++a) map[a] = '?';
    va_list ap;
    va_start(ap,n);
    char ch = '?';
    while(n != -2)
    {
        if(n==-1)
            ch = va_arg(ap, int);
        else
        {
            int p = va_arg(ap, int)&3;
            map[n*4+p] = ch;
        }
        n = va_arg(ap, int);
    }
    va_end(ap);
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c = s.outmap[y*16+x].tile;
            unsigned char p = s.outmap[y*16+x].palette;
            s.ascii[y*32+   x] = map[c*4+p];
            s.ascii[y*32+16+x] = ' ';
        }
}
